/* Copyright (c) 2013 tonomu
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

enum version {
  MAJOR_VERSION = 1,
  MINOR_VERSION = 0,
  MODIFY_VERSION = 0
};

enum { FILENAME_LEN = 256 };

struct header_st {
  int format;
  int total;
  int tpqn;
};

struct block_st {
  long time;
  unsigned char *event;
};

struct track_st {
  struct block_st *block;
};

struct midi_st {
  struct header_st header;
  struct track_st *track;
};

struct arg_st {
  int compile;
  int format;
  int inq;
  int outq;
  int version;
  int help;
  char infile[FILENAME_LEN];
  char outfile[FILENAME_LEN];
};

struct beat_st {
  int num;
  int den;
};

struct rbar_st {
  long tick;
  struct beat_st beat;
};

struct revt_st {
  unsigned char *in;
  unsigned char *out;
  unsigned char *top;
  long dur;
  long cnt;
  long eol;
  int ch;
};

struct rmbt_st {
  int eof;

  int eot;
  int trk;

  int eob;
  int blk;
  struct block_st *block;

  long max;
  long lim;
  struct rbar_st *bar;

  long sep;
  long row;
  struct revt_st buf;

  long mbt[3];
};

struct wbar_st {
  long tick;
  int key;
  struct beat_st beat;
};

struct wmbt_st {
  int trk;

  long max;
  long lim;
  struct wbar_st *bar;

  int tpqn;

  int ch;

  long mbt[3];
};

enum error_code {
  STATUS_SUCCESS = 0,
  E_A_NO_INFO = -1,

  E_M_INPUT_TPQN = -2,
  E_M_CLOSE_INPUT = -3,
  E_M_GCD_VALUE = -4,
  E_M_FORMAT_ZERO = -5,
  E_M_CLOSE_OUTPUT = -6,
  E_M_FLAG_FORMAT = -7,
  E_M_ARGUMENT_END = -8,
  E_M_FLAG_TYPE = -9,
  E_M_ASSIGN_FNAME = -10,
  E_M_SET_FNAME = -11,
  E_M_FNAME_LEN = -12,
  E_M_ASSIGN_DATA = -13,
  E_M_SET_TPQN = -14,
  E_M_TPQN_VALUE = -15,
  E_M_TPQN_DIGIT = -16,
  E_M_TPQN_ZERO = -17,
  E_M_TICK_VALUE = -18,
  E_M_INPUT_FNAME = -19,
  E_M_OPEN_INPUT = -20,
  E_M_OPEN_OUTPUT = -21,
  E_M_RFU_1 = -22,
  E_M_RFU_2 = -23,
  E_M_RFU_3 = -24,

  E_G_TRACK_MALLOC = -25,
  E_G_MTHD_DATA = -26,
  E_G_HEADER_SIZE = -27,
  E_G_SMF_FORMAT = -28,
  E_G_MTRK_DATA = -29,
  E_G_TRACK_LEN = -30,
  E_G_BLOCK_MALLOC = -31,
  E_G_TRACK_DATA = -32,
  E_G_TRACK_SIZE = -33,
  E_G_BLOCK_REALLOC = -34,
  E_G_GET_CHAR = -35,
  E_G_TIME_VALUE = -36,
  E_G_LEN_FORMAT = -37,
  E_G_SET_RNG = -38,
  E_G_EVENT_MALLOC = -39,
  E_G_EVENT_TYPE = -40,
  E_G_RFU_1 = -41,
  E_G_RFU_2 = -42,
  E_G_RFU_3 = -43,

  E_W_TPQN_VALUE = -44,
  E_W_TPQN_DATA = -45,
  E_W_WRITE_HEADER = -46,
  E_W_BAR_MALLOC = -47,
  E_W_TRACK_COM = -48,
  E_W_BAR_VALUE = -49,
  E_W_BAR_REALLOC = -50,
  E_W_MEAS_TICK = -51,
  E_W_WRITE_MBT = -52,
  E_W_EVENT_TYPE = -53,
  E_W_WRITE_INS = -54,
  E_W_WRITE_DATA = -55,
  E_W_BEND_DATA = -56,
  E_W_WRITE_TONE = -57,
  E_W_LEN_FORMAT = -58,
  E_W_WRITE_COM = -59,
  E_W_WRITE_TEXT = -60,
  E_W_WRITE_TEMPO = -61,
  E_W_WRITE_TIMESIG = -62,
  E_W_TIMESIG_VALUE = -63,
  E_W_WRITE_KEY = -64,
  E_W_BEAT_DATA = -65,
  E_W_RFU_1 = -66,
  E_W_RFU_2 = -67,
  E_W_RFU_3 = -68,

  E_R_LINE_DATA = -69,
  E_R_TPQN_VALUE = -70,
  E_R_TPQN_DATA = -71,
  E_R_TRACK_MALLOC = -72,
  E_R_INFBLOCK_MALLOC = -73,
  E_R_INFBUF_MALLOC = -74,
  E_R_INFBAR_MALLOC = -75,
  E_R_ROW_VALUE = -76,
  E_R_STACK_CHAR = -77,
  E_R_SET_SEPARATOR = -78,
  E_R_EOF_POS = -79,
  E_R_RET_POS = -80,
  E_R_LINE_VALUE = -81,
  E_R_INFBUF_REALLOC = -82,
  E_R_BLOCK_COUNT = -83,
  E_R_INFBLOCK_REALLOC = -84,
  E_R_DELTA_TICK = -85,
  E_R_REMAIN_TIMEDATA = -86,
  E_R_TIME_CHAR = -87,
  E_R_MBT_CHAR = -88,
  E_R_BEAT_VALUE = -89,
  E_R_MAKE_MBT = -90,
  E_R_BAR_COUNT = -91,
  E_R_BAR_REALLOC = -92,
  E_R_MEAS_TICK = -93,
  E_R_MINUS_TICK = -94,
  E_R_EVENT_DATA = -95,
  E_R_DATA_LEN = -96,
  E_R_DUR_VALUE = -97,
  E_R_REMAIN_EVENT = -98,
  E_R_FXMG_VALUE = -99,
  E_R_CHAIN_DATA = -100,
  E_R_LEN_DATA = -101,
  E_R_DATA_VALUE = -102,
  E_R_PREFIX_TYPE = -103,
  E_R_HEX_VALUE = -104,
  E_R_HEX_DATA = -105,
  E_R_EVENT_BUF = -106,
  E_R_SET_CHANNEL = -107,
  E_R_FIX_CHANNEL = -108,
  E_R_CHANNEL_VALUE = -109,
  E_R_HEX_CHARACTER = -110,
  E_R_TONE_LOWER = -111,
  E_R_TONE_NUMBER = -112,
  E_R_TONE_UPPER = -113,
  E_R_TONE_NAME = -114,
  E_R_PITCH_VALUE = -115,
  E_R_TEMPO_VALUE = -116,
  E_R_TEMPO_FAULT = -117,
  E_R_TEMPO_MIN = -118,
  E_R_STR_CLOSE = -119,
  E_R_TEXT_LEN = -120,
  E_R_MINUS_VALUE = -121,
  E_R_BEAT_DATA = -122,
  E_R_BEAT_UPPER = -123,
  E_R_NUM_VALUE = -124,
  E_R_METER_CHANGE = -125,
  E_R_SET_BEAT = -126,
  E_R_EVENT_MALLOC = -127,
  E_R_MUTE_DUR = -128,
  E_R_TRACK_COUNT = -129,
  E_R_TRACK_REALLOC = -130,
  E_R_BLOCK_MALLOC = -131,
  E_R_RFU_1 = -132,
  E_R_RFU_2 = -133,
  E_R_RFU_3 = -134,

  E_P_TRACK_COUNT = -135,
  E_P_TPQN_VALUE = -136,
  E_P_OUT_MTHD = -137,
  E_P_OUT_HEADSIZE = -138,
  E_P_FORMAT_TYPE = -139,
  E_P_OUT_FORMAT = -140,
  E_P_OUT_TRACK = -141,
  E_P_OUT_TPQN = -142,
  E_P_POP_START = -143,
  E_P_POP_END = -144,
  E_P_SET_START = -145,
  E_P_SET_END = -146,
  E_P_OUT_MTRK = -147,
  E_P_OUT_TEMPSIZE = -148,
  E_P_TRACK_SIZE = -149,
  E_P_OUT_DATASIZE = -150,
  E_P_DELTA_TICK = -151,
  E_P_OUT_TICK = -152,
  E_P_OUT_EVENTDATA = -153,
  E_P_EVENT_TYPE = -154,
  E_P_DATA_CHAIN = -155,
  E_P_DATA_LEN = -156,
  E_P_RFU_1 = -157,
  E_P_RFU_2 = -158,
  E_P_RFU_3 = -159
};

int Get_Data(FILE *ifp, struct midi_st *midi);
int Write_Data(struct midi_st *midi, FILE *ofp);
int Read_Data(FILE *ifp, struct midi_st *midi);
int Put_Data(struct midi_st *midi, FILE *ofp);
int Print_Error(int status);
